"""
Classes and methods for parsing and using digital elevation models in DTED format.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
import os
import pathlib
import struct

import numpy

from sarpy.io.DEM.DEM import DEMList, DEMInterpolator
from sarpy.io.DEM.utils import argument_validation
from sarpy.io.DEM.geoid import GeoidHeight
from sarpy.io.general.base import SarpyIOError

logger = logging.getLogger(__name__)


#######
# module variables
_SUPPORTED_DTED_FILE_TYPES = {
    'DTED1': {'fext': '.dt1'},
    'DTED2': {'fext': '.dt2'},
    'SRTM1': {'fext': '.dt1'},
    'SRTM2': {'fext': '.dt2'},
    'SRTM2F': {'fext': '.dt2'}}


def get_default_prioritization():
    """
    Gets the default prioritization of the DTED types.

    Returns
    -------
    List[str]
    """

    # TODO: what should this actually be?
    return 'DTED2', 'DTED1', 'SRTM2F', 'SRTM2', 'SRTM1'


def get_lat_lon_box(lats, lons):
    """
    Gets the lat/lon bounding box, as appropriate.

    Parameters
    ----------
    lats : numpy.ndarray|list|tuple|float|int
    lons : numpy.ndarray|list|tuple|float|int

    Returns
    -------
    numpy.ndarray
    """

    def get_min_max(inp, lon=False):
        if isinstance(inp, (int, float, numpy.number)):
            return inp, inp
        else:
            min_val, max_val = numpy.min(inp), numpy.max(inp)
            if not lon:
                return numpy.min(inp), numpy.max(inp)
            # check for 180/-180 crossing
            if not (min_val < -90 and max_val > 90):
                return min_val, max_val
            inp = numpy.array(inp).flatten()
            min_val = numpy.min(inp[inp >= 0])
            max_val = numpy.max(inp[inp <= 0])
            return min_val, max_val

    out = numpy.zeros((4, ), dtype='float64')
    out[:2] = get_min_max(lats)
    out[2:] = get_min_max(lons, lon=True)
    return out


class DTEDList(DEMList):
    """
    The DEM directory structure is assumed to look like the following:

        * For DTED<1,2>: `<root_dir>/dted/<1, 2>/<lon_string>/<lat_string>.dtd`

        * For SRTM<1,2,2F>: `<root_dir>/srtm/<1, 2, 2f>/<lon_string>/<lat_string>.dt<1,2,2>`

    Here `<lat_string>` corresponds to a string of the form `X##`, where
    `X` is one of 'N' or 'S', and `##` is the zero-padded formatted string for
    the integer value `floor(lat)`.

    Similarly, `<lon_string>` corresponds to a string of the form `Y##`, where
    `Y` is one of 'E' or 'W', and `###` is the zero-padded formatted string for
    the integer value `floor(lon)`.

    <lon_string>, <lat_string> corresponds to the origin in the lower left corner
    of the DEM tile.
    """

    __slots__ = ('_root_dir', '_missing_error')

    def __init__(self, root_directory, missing_error=False):
        """

        Parameters
        ----------
        root_directory : str
        missing_error : bool
            Raise an exception when DTED files are missing?
        """

        self._root_dir = root_directory
        self._missing_error = missing_error

    @property
    def root_dir(self):
        """
        str: the root directory
        """

        return self._root_dir

    def _get_directory_stem(self, dem_type):
        if dem_type.startswith('DTED'):
            return os.path.join(self._root_dir, dem_type[:4].lower(), dem_type[-1])
        elif dem_type.startswith('SRTM'):
            return os.path.join(self._root_dir, dem_type[:4].lower(), dem_type[4:].lower())
        else:
            raise ValueError('Unhandled dem_type {}'.format(dem_type))

    def _get_file_list(self, lat_lon_list, dem_type):
        """
        Helper method for getting the file list for a specified type.

        Parameters
        ----------
        lat_lon_list : list
        dem_type : str

        Returns
        -------
        List[str]
        """

        def get_box(la, lo):
            x = 'n' if la >= 0 else 's'
            y = 'e' if lo >= 0 else 'w'
            return '{0:s}{1:03d}'.format(y, abs(lo)), '{0:s}{1:02d}'.format(x, abs(la))

        # get the directory search stem
        dstem = self._get_directory_stem(dem_type)
        if not os.path.isdir(dstem):
            return  # nothing to be done

        # get file extension
        fext = _SUPPORTED_DTED_FILE_TYPES[dem_type]['fext']

        for entry in lat_lon_list:
            if entry[2] is not None:
                # we already found the file
                continue
            lonstr, latstr = get_box(entry[0], entry[1])
            fil = os.path.join(dstem, lonstr, latstr + fext)
            if os.path.isfile(fil):
                entry[2] = fil

    def get_file_list(self, lat_lon_box, dem_type=None):
        """
        Get the file list required for the given coordinates.

        Parameters
        ----------
        lat_lon_box : numpy.ndarray|list|tuple
            The bounding box of the form `[lat min, lat max, lon min, lon max]`.
        dem_type : None|str|List[str]
            The prioritized list of dem types to check. If `None`, then
            :func:`get_default_prioritization` is used. Each entry must be one
            of ("DTED1", "DTED2", "SRTM1", "SRTM2", "SRTM2F")

        Returns
        -------
        List[str]
        """

        # let's construct the list of lats that we must match
        lat_start = int(numpy.floor(lat_lon_box[0]))
        lat_end = int(numpy.ceil(lat_lon_box[1]))
        if (lat_start > lat_end) or (lat_start < -90) or (lat_start > 90):
            raise ValueError('Got malformed latitude in bounding box {}'.format(lat_lon_box))
        if lat_start == lat_end:
            lat_list = [lat_start, ]
        else:
            lat_list = list(range(lat_start, lat_end, 1))

        # let's construct the list of lons that we must match.
        lon_start = int(numpy.floor(lat_lon_box[2]))
        lon_end = int(numpy.ceil(lat_lon_box[3]))
        if (lon_start < -180) or (lon_end < -180) or (lon_start > 180) or (lon_end > 180):
            raise ValueError('Got malformed longitude in bounding box {}'.format(lat_lon_box))
        if lon_start > lon_end:
            if not (lon_end < 0 < lon_start):
                # this is assumed to NOT be a 180/-180 boundary crossing
                raise ValueError(
                    'We have minimum longitude greater than maximum longitude {}'.format(lat_lon_box))
            else:
                # we have a 180/-180 boundary crossing
                lon_list = list(range(lon_start, 180, 1)) + list(range(-180, lon_end, 1))
        else:
            if lon_start == lon_end:
                lon_list = [lon_start, ]
            else:
                lon_list = list(range(lon_start, lon_end, 1))

        # construct our workspace
        lat_lon_list = []
        for corner_lat in lat_list:
            for corner_lon in lon_list:
                lat_lon_list.append([corner_lat, corner_lon, None])

        # validate the dem types list
        if dem_type is None:
            dem_type = get_default_prioritization()
        elif isinstance(dem_type, str):
            dem_type = [dem_type, ]
        # loop over the prioritized list of types and check
        for entry in dem_type:
            if not isinstance(entry, str):
                raise TypeError(
                    'Got entry {} of dem_type, this is required to be of string type'.format(entry))
            # validate dem_type options
            this_entry = entry.upper()
            if this_entry not in _SUPPORTED_DTED_FILE_TYPES:
                raise ValueError(
                    'Got dem_type {}, but it must be one of the supported '
                    'types {}'.format(entry, list(_SUPPORTED_DTED_FILE_TYPES.keys())))
            self._get_file_list(lat_lon_list, this_entry)  # NB: this modifies lat_lon_list in place

        # extract files and warn about missing entries
        files = []
        missing_boxes = []
        for entry in lat_lon_list:
            if entry[2] is not None:
                files.append(entry[2])
            else:
                missing_boxes.append('({}, {})'.format(entry[0], entry[1]))

        if len(missing_boxes) > 0:
            msg = 'Missing expected DEM files for squares with lower left lat/lon corner {}'.format(missing_boxes)
            if self._missing_error:
                raise ValueError(msg)
            else:
                logger.warning(
                    msg + '\n\tThis should result in the assumption that the altitude in\n\t'
                          'that section is given by Mean Sea Level.')
        return files


class DTEDReader(object):
    """
    Reader/interpreter for DTED files, generally expected to be a helper class.
    As such, some implementation choices have been made for computational efficiency,
    and not user convenience.
    """

    __slots__ = ('_file_name', '_origin', '_spacing', '_bounding_box', '_shape', '_mem_map')

    def __init__(self, file_name):
        self._file_name = file_name

        with open(self._file_name, 'rb') as fi:
            # NB: DTED is always big-endian
            # the first 80 characters are header
            # characters 80:728 are data set identification record
            # characters 728:3428 are accuracy record
            # the remainder is data records, but DTED is not quite a raster
            header = struct.unpack('>80s', fi.read(80))[0].decode('utf-8')

        if header[:3] != 'UHL':
            raise SarpyIOError('File {} does not appear to be a DTED file.'.format(self._file_name))

        lon = float(header[4:7]) + float(header[7:9])/60. + float(header[9:11])/3600.
        lon = -lon if header[11] == 'W' else lon
        lat = float(header[12:15]) + float(header[15:17])/60. + float(header[17:19])/3600.
        lat = -lat if header[19] == 'S' else lat

        self._origin = numpy.array([lon, lat], dtype=numpy.float64)
        self._spacing = numpy.array([float(header[20:24]), float(header[24:28])], dtype=numpy.float64)/36000.
        self._shape = numpy.array([int(header[47:51]), int(header[51:55])], dtype=numpy.int64)
        self._bounding_box = numpy.zeros((4, ), dtype=numpy.float64)
        self._bounding_box[0::2] = self._origin
        self._bounding_box[1::2] = self._origin + self._spacing*(self._shape - 1)

        # starting at 3428, the rest of the file is data records, but not quite a raster
        #   each "row" is a data record with 8 extra bytes at the beginning,
        #   and 4 extra (checksum) at the end - look to MIL-PRF-89020B for an explanation
        # To enable memory map usage, we will spoof it as a raster and adjust column indices
        shp = (int(self._shape[0]), int(self._shape[1]) + 6)
        self._mem_map = numpy.memmap(self._file_name,
                                     dtype=numpy.dtype('>u2'),
                                     mode='r',
                                     offset=3428,
                                     shape=shp)

    @property
    def origin(self):
        """
        numpy.ndarray: The origin of this DTED, of the form `[longitude, latitude]`.
        """

        return numpy.copy(self._origin)

    @property
    def bounding_box(self):
        """
        numpy.ndarray: The bounding box of the form
        `[longitude min, longitude max, latitude min, latitude max]`.
        """

        return numpy.copy(self._bounding_box)

    def __getitem__(self, item):
        def new_col_int(val, begin):
            if val is None:
                if begin:
                    return 4
                else:
                    return -2
            return val + 4 if val >= 0 else val - 2

        # we need to manipulate in the second dimension
        if isinstance(item, tuple):
            if len(item) > 2:
                raise ValueError('Cannot slice on more than 2 dimensions')
            it = item[1]
            if isinstance(it, int):
                it1 = new_col_int(it, True)
            elif isinstance(it, slice):
                start = new_col_int(it.start, True)
                stop = new_col_int(it.stop, False)
                it1 = slice(start, stop, it.step)
            elif isinstance(item[1], numpy.ndarray):
                it1 = numpy.copy(item[1])
                it1[it1 >= 0] += 4
                it1[it1 < 0] -= 2
            else:
                raise ValueError('Cannot slice using {}'.format(type(item[1])))
            data = self._mem_map.__getitem__((item[0], it1))
        else:
            data = self._mem_map[item, 4:-2]
        return self._repair_values(data)

    @staticmethod
    def _repair_values(elevations):
        """
        Convert 2-bytes signed magnitude to twos complement.

        Parameters
        ----------
        elevations : numpy.ndarray

        Returns
        -------
        numpy.ndarray

        Notes
        -----
        Per MIL-PRF-89020B Section 3.11.1:

            All elevation values are signed magnitude binary integers, right justified,
            16 bits (2 bytes). The sign bit is in the high order position.

        """
        out = (elevations & 0x7f_ff).astype(numpy.int16)
        out *= (-1) ** ((elevations & 0x80_00) != 0)
        return out

    def _linear(self, ix, dx, iy, dy):
        # type: (numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray) -> numpy.ndarray
        a = (1 - dx) * self._lookup_elevation(ix, iy) + dx * self._lookup_elevation(ix + 1, iy)
        b = (1 - dx) * self._lookup_elevation(ix, iy+1) + dx * self._lookup_elevation(ix+1, iy+1)
        return (1 - dy) * a + dy * b

    def _lookup_elevation(self, ix, iy):
        # type: (numpy.ndarray, numpy.ndarray) -> numpy.ndarray
        t_ix = numpy.copy(ix)
        t_ix[t_ix >= self._shape[0]] = self._shape[0] - 1
        t_ix[t_ix < 0] = 0

        # adjust iy to account for 8 extra bytes at the beginning of each column
        t_iy = iy + 4
        t_iy[t_iy >= self._shape[1]+4] = self._shape[1] + 3
        t_iy[t_iy < 4] = 4

        return self._repair_values(self._mem_map[t_ix, t_iy])

    def in_bounds(self, lat, lon):
        """
        Determine which of the given points are inside the extent of this DTED.

        Parameters
        ----------
        lat : numpy.ndarray
        lon : numpy.ndarray

        Returns
        -------
        numpy.ndarray
            boolean array of the same shape as lat/lon
        """

        return (lon >= self._bounding_box[0]) & (lon <= self._bounding_box[1]) & \
               (lat >= self._bounding_box[2]) & (lat <= self._bounding_box[3])

    def _get_elevation(self, lat, lon):
        # type: (numpy.ndarray, numpy.ndarray) -> numpy.ndarray

        # we implicitly require that lat/lon make sense and are contained in this DTED

        # get indices
        fx = (lon - self._origin[0])/self._spacing[0]
        fy = (lat - self._origin[1])/self._spacing[1]

        # get integer indices via floor
        ix = numpy.asarray(numpy.floor(fx), dtype=numpy.int32)
        iy = numpy.asarray(numpy.floor(fy), dtype=numpy.int32)
        return self._linear(ix, fx-ix, iy, fy-iy)

    def get_elevation(self, lat, lon, block_size=50000):
        """
        Interpolate the elevation values for lat/lon. This is relative to the EGM96
        geoid by DTED specification.

        Parameters
        ----------
        lat : numpy.ndarray
        lon : numpy.ndarray
        block_size : None|int
            If `None`, then the entire calculation will proceed as a single block.
            Otherwise, block processing using blocks of the given size will be used.
            The minimum value used for this is 50,000, and any smaller value will be
            replaced with 50,000. Default is 50,000.

        Returns
        -------
        numpy.ndarray
            Elevation values of the same shape as lat/lon.
        """

        o_shape, lat, lon = argument_validation(lat, lon)

        out = numpy.full(lat.shape, numpy.nan, dtype=numpy.float64)
        if block_size is None:
            boolc = self.in_bounds(lat, lon)
            if numpy.any(boolc):
                out[boolc] = self._get_elevation(lat[boolc], lon[boolc])
        else:
            block_size = min(50000, int(block_size))
            start_block = 0
            while start_block < lat.size:
                end_block = min(lat.size, start_block + block_size)
                lat1 = lat[start_block:end_block]
                lon1 = lon[start_block:end_block]
                boolc = self.in_bounds(lat1, lon1)
                out1 = numpy.full(lat1.shape, numpy.nan, dtype=numpy.float64)
                out1[boolc] = self._get_elevation(lat1[boolc], lon[boolc])
                out[start_block:end_block] = out1
                start_block = end_block

        if o_shape == ():
            return float(out[0])
        else:
            return numpy.reshape(out, o_shape)

    def _find_overlap(self, lat_lon_box):
        """
        Gets the overlap slice argument for the lat/lon bounding box.

        Parameters
        ----------
        lat_lon_box : None|numpy.ndarray
            Of the form `[

        Returns
        -------
        (slice, slice)
        """

        if lat_lon_box is None:
            first_row = 0
            last_row = self._shape[0]
            first_col = 0
            last_col = self._shape[1]
        else:
            first_row = (lat_lon_box[2] - self._origin[0])/self._spacing[0]
            last_row = (lat_lon_box[3] - self._origin[0])/self._spacing[0]
            first_col = (lat_lon_box[0] - self._origin[1])/self._spacing[1]
            last_col = (lat_lon_box[1] - self._origin[1])/self._spacing[1]
            if first_row > self._shape[0] or last_row < 0 or first_col > self._shape[1] or last_col < 0:
                return None
            first_row = int(max(0, numpy.floor(first_row)))
            last_row = int(min(self._shape[0], numpy.ceil(last_row)))
            first_col = int(max(0, numpy.floor(first_col)))
            last_col = int(min(self._shape[1], numpy.ceil(last_col)))
        return slice(first_row, last_row, 1), slice(first_col, last_col, 1)

    def get_max(self, lat_lon_box=None):
        """
        Gets the maximum observed DEM value, possibly contained in the given
        rectangular area of interest.

        Parameters
        ----------
        lat_lon_box : None|numpy.ndarray
            None or of the form `[lat min, lat max, lon min, lon max]`.

        Returns
        -------
        float|None
        """

        arg = self._find_overlap(lat_lon_box)
        if arg is None:
            return None
        return numpy.max(self.__getitem__(arg))

    def get_min(self, lat_lon_box=None):
        """
        Gets the minimum observed DEM value, possibly contained in the given
        rectangular area of interest.

        Parameters
        ----------
        lat_lon_box : None|numpy.ndarray
            None or of the form `[lat min, lat max, lon min, lon max]`.

        Returns
        -------
        float|None
        """

        arg = self._find_overlap(lat_lon_box)
        if arg is None:
            return None
        return numpy.min(self.__getitem__(arg))


class DTEDInterpolator(DEMInterpolator):
    """
    DEM Interpolator using DTED/SRTM files for the DEM information.
    """

    __slots__ = ('_readers', '_geoid', '_ref_geoid')

    def __init__(self, files, geoid_file, lat_lon_box=None):
        if isinstance(files, str):
            files = [files, ]
        # get a reader object for each file
        self._readers = [DTEDReader(fil) for fil in files]

        # get the geoid object - we should prefer egm96 .pgm files, since that's the DTED spec
        #   in reality, it makes very little difference, though
        if isinstance(geoid_file, (str, pathlib.Path)):
            if os.path.isdir(geoid_file):
                geoid_file = GeoidHeight.from_directory(geoid_file, search_files=('egm96-5.pgm', 'egm96-15.pgm'))
            else:
                geoid_file = GeoidHeight(geoid_file)
        if not isinstance(geoid_file, GeoidHeight):
            raise TypeError(
                'geoid_file is expected to be the path where one of the standard '
                'egm .pgm files can be found, or an instance of GeoidHeight reader. '
                'Got {}'.format(type(geoid_file)))
        self._geoid = geoid_file

        self._lat_lon_box = lat_lon_box

        if len(self._readers) == 0:
            self._ref_geoid = 0
        else:
            # use the origin of the first reader
            ref_point = self._readers[0].origin
            self._ref_geoid = float(self._geoid.get(ref_point[1], ref_point[0]))
        self._max_geoid = None
        self._min_geoid = None

    @classmethod
    def from_coords_and_list(cls, lat_lon_box, dted_list, dem_type=None, geoid_file=None):
        """
        Construct a `DTEDInterpolator` from a coordinate collection and `DTEDList` object.

        .. Note:: This depends on using :func:`DTEDList.get_file_list`
            to get the relevant file list.

        Parameters
        ----------
        lat_lon_box : numpy.ndarray|list|tuple
            Of the form `[lat min, lat max, lon min, lon max]`.
        dted_list : DTEDList|str
            The dted list object or root directory
        dem_type : None|str|List[str]
            The DEM type or list of DEM types in order of priority.
        geoid_file : None|str|GeoidHeight
            The `GeoidHeight` object, an egm file name, or root directory containing
            one of the egm files in the sub-directory "geoid". If `None`, then default
            to the root directory of `dted_list`.

        Returns
        -------
        DTEDInterpolator
        """

        if isinstance(dted_list, str):
            dted_list = DTEDList(dted_list)
        if not isinstance(dted_list, DTEDList):
            raise ValueError(
                'dted_list os required to be a path (directory) or DTEDList instance.')

        # default the geoid argument to the root directory of the dted_list
        if geoid_file is None:
            geoid_file = dted_list.root_dir

        return cls(dted_list.get_file_list(lat_lon_box, dem_type=dem_type), geoid_file, lat_lon_box=lat_lon_box)

    @classmethod
    def from_reference_point(cls, ref_point, dted_list, dem_type=None, geoid_file=None, pad_value=0.1):
        """
        Construct a DTEDInterpolator object by padding around the reference point by
        `pad_value` latitude degrees (1 degree ~ 111 km or 69 miles).

        .. Note:: The degeneracy at the poles is not handled, because DTED are not
            defined there anyways.

        Parameters
        ----------
        ref_point : numpy.ndarray|list|tuple
            This is assumed to be of the form `[lat, lon, ...]`, and entries
            beyond the first two are ignored.
        dted_list : DTEDList|str
            The dted list object or root directory
        dem_type : None|str|List[str]
            The DEM type or list of DEM types in order of priority.
        geoid_file : None|str|GeoidHeight
            The `GeoidHeight` object, an egm file name, or root directory containing
            one of the egm files in the sub-directory "geoid". If `None`, then default
            to the root directory of `dted_list`.
        pad_value : float
            The degree value to pad by.

        Returns
        -------
        DTEDInterpolator
        """

        pad_value = float(pad_value)
        if pad_value > 0.5:
            pad_value = 0.5
        if pad_value < 0.05:
            pad_value = 0.05
        lat_diff = pad_value
        lat_max = min(ref_point[0] + lat_diff, 90)
        lat_min = max(ref_point[0] - lat_diff, -90)

        lon_diff = min(15, lat_diff/(numpy.sin(numpy.deg2rad(ref_point[0]))))
        lon_max = ref_point[1] + lon_diff
        if lon_max > 180:
            lon_max -= 360
        lon_min = ref_point[1] - lon_diff
        if lon_min < -180:
            lon_min += 360

        return cls.from_coords_and_list(
            [lat_min, lat_max, lon_min, lon_max], dted_list, dem_type=dem_type, geoid_file=geoid_file)

    @property
    def geoid(self):  # type: () -> GeoidHeight
        """
        GeoidHeight: Get the geoid height calculator
        """

        return self._geoid

    def _get_elevation_geoid_from_reader(self, reader, lat, lon):
        mask = reader.in_bounds(lat, lon)
        values = numpy.full(lat.shape, numpy.nan, dtype=numpy.float64)
        if numpy.any(mask):
            # noinspection PyProtectedMember
            values[mask] = reader._get_elevation(lat[mask], lon[mask])
        return mask, values

    def _get_elevation_geoid(self, lat, lon):
        out = numpy.full(lat.shape, numpy.nan, dtype=numpy.float64)
        remaining = numpy.ones(lat.shape, dtype=numpy.bool_)
        for reader in self._readers:
            if not numpy.any(remaining):
                break
            mask, values = self._get_elevation_geoid_from_reader(reader, lat[remaining], lon[remaining])
            if numpy.any(mask):
                work_mask = numpy.copy(remaining)
                work_mask[remaining] = mask  # mask as a subset of remaining
                out[work_mask] = values[mask]
                remaining[work_mask] = False
        return out

    def get_elevation_hae(self, lat, lon, block_size=50000):
        """
        Get the elevation value relative to the WGS-84 ellipsoid.

        .. Note:: DTED elevation is relative to the egm96 geoid, and we are simply adding
            values determined by a geoid calculator. Using a the egm2008 model will result
            in only minor differences.

        Parameters
        ----------
        lat : numpy.ndarray|list|tuple|int|float
        lon : numpy.ndarray|list|tuple|int|float
        block_size : None|int
            If `None`, then the entire calculation will proceed as a single block.
            Otherwise, block processing using blocks of the given size will be used.
            A minimum value of 50000 will be enforced here.

        Returns
        -------
        numpy.ndarray
            The elevation relative to the WGS-84 ellipsoid.
        """

        return self.get_elevation_geoid(lat, lon, block_size=block_size) + \
            self._geoid.get(lat, lon, block_size=block_size)

    def get_elevation_geoid(self, lat, lon, block_size=50000):
        """
        Get the elevation value relative to the geoid.

        .. Note:: DTED elevation is relative to the egm96 geoid, though using the egm2008
            model will result in only minor differences.

        Parameters
        ----------
        lat : numpy.ndarray|list|tuple|int|float
        lon : numpy.ndarray|list|tuple|int|float
        block_size : None|int
            If `None`, then the entire calculation will proceed as a single block.
            Otherwise, block processing using blocks of the given size will be used.
            The minimum value used for this is 50000, and any smaller value will be
            replaced with 50000. Default is 50000.

        Returns
        -------
        numpy.ndarray
            the elevation relative to the geoid
        """

        o_shape, lat, lon = argument_validation(lat, lon)

        if block_size is None:
            out = self._get_elevation_geoid(lat, lon)
        else:
            block_size = min(50000, int(block_size))
            out = numpy.full(lat.shape, numpy.nan, dtype=numpy.float64)
            start_block = 0
            while start_block < lat.size:
                end_block = min(lat.size, start_block+block_size)
                out[start_block:end_block] = self._get_elevation_geoid(
                    lat[start_block:end_block], lon[start_block:end_block])
                start_block = end_block
        out[numpy.isnan(out)] = 0.0  # set missing values to geoid=0 (MSL)

        if o_shape == ():
            return float(out[0])
        else:
            return numpy.reshape(out, o_shape)

    def _get_ref_geoid(self, lat_lon_box):
        if lat_lon_box is None:
            # use the origin of the first reader
            ref_point = self._readers[0].origin
            return float(self._geoid.get(ref_point[1], ref_point[0]))
        else:
            return float(self._geoid.get(lat_lon_box[0], lat_lon_box[2]))

    def get_max_hae(self, lat_lon_box=None):
        return self.get_max_geoid(lat_lon_box=lat_lon_box) + self._get_ref_geoid(lat_lon_box)

    def get_min_hae(self, lat_lon_box=None):
        return self.get_min_geoid(lat_lon_box=lat_lon_box) + self._get_ref_geoid(lat_lon_box)

    def get_max_geoid(self, lat_lon_box=None):
        obs_maxes = [reader.get_max(lat_lon_box=lat_lon_box) for reader in self._readers]
        obs_maxes = [value for value in obs_maxes if value is not None]

        if not obs_maxes:
            return 0  # Assume missing area to be MSL

        return float(max(obs_maxes))

    def get_min_geoid(self, lat_lon_box=None):
        obs_mins = [reader.get_min(lat_lon_box=lat_lon_box) for reader in self._readers]
        obs_mins = [value for value in obs_mins if value is not None]

        if not obs_mins:
            return 0  # Assume missing area to be MSL

        return float(min(obs_mins))




"""
The GridType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from typing import Union, Optional, Dict, Tuple

import numpy
from numpy.linalg import norm
from scipy.constants import speed_of_light

from sarpy.processing.sicd.windows import general_hamming, taylor, kaiser, find_half_power, \
    get_hamming_broadening_factor
from sarpy.io.xml.base import Serializable, ParametersCollection, find_first_child
from sarpy.io.xml.descriptors import StringDescriptor, StringEnumDescriptor, \
    FloatDescriptor, FloatArrayDescriptor, IntegerEnumDescriptor, \
    SerializableDescriptor, UnitVectorDescriptor, ParametersDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import XYZType, Poly2DType
from .utils import _get_center_frequency


logger = logging.getLogger(__name__)


# module variable
DEFAULT_WEIGHT_SIZE = 512
"""
int: the default size when generating WgtFunct from a named WgtType.
"""


class WgtTypeType(Serializable):
    """
    The weight type parameters of the direction parameters.
    """

    _fields = ('WindowName', 'Parameters')
    _required = ('WindowName',)
    _collections_tags = {'Parameters': {'array': False, 'child_tag': 'Parameter'}}
    # descriptors
    WindowName = StringDescriptor(
        'WindowName', _required, strict=DEFAULT_STRICT,
        docstring='Type of aperture weighting applied in the spatial frequency domain (Krow) to yield '
                  'the impulse response in the row direction. '
                  '*Example values - "UNIFORM", "TAYLOR", "UNKNOWN", "HAMMING"*')  # type: str
    Parameters = ParametersDescriptor(
        'Parameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Free form parameters list.')  # type: ParametersCollection

    def __init__(
            self,
            WindowName: str = None,
            Parameters: Union[ParametersCollection, Dict] = None,
            **kwargs):
        """

        Parameters
        ----------
        WindowName : str
        Parameters : ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.WindowName = WindowName
        self.Parameters = Parameters
        super(WgtTypeType, self).__init__(**kwargs)

    def get_parameter_value(
            self,
            param_name: str,
            default: Optional[str] = None) -> Optional[str]:
        """
        Gets the value (first value found) associated with a given parameter name.
        Returns `default` if not found.

        Parameters
        ----------
        param_name : str
            the parameter name for which to search.
        default : None|str
            the default value to return if lookup fails.

        Returns
        -------
        str
            the associated parameter value, or `default`.
        """

        if self.Parameters is None:
            return default
        the_dict = self.Parameters.get_collection()
        if len(the_dict) == 0:
            return default
        if param_name is None:
            # get the first value - this is dumb, but appears a use case. Leaving undocumented.
            return list(the_dict.values())[0]
        return the_dict.get(param_name, default)

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        win_key = cls._child_xml_ns_key.get('WindowName', ns_key)
        win_name = find_first_child(node, 'WindowName', xml_ns, win_key)
        if win_name is None:
            # SICD 0.4 standard compliance, this could just be a space delimited string of the form
            #   "<WindowName> <name1>=<value1> <name2>=<value2> ..."
            if kwargs is None:
                kwargs = {}
            values = node.text.strip().split()
            kwargs['WindowName'] = values[0]
            params = {}
            for entry in values[1:]:
                try:
                    name, val = entry.split('=')
                    params[name] = val
                except ValueError:
                    continue
            kwargs['Parameters'] = params
            return cls.from_dict(kwargs)
        else:
            return super(WgtTypeType, cls).from_node(node, xml_ns, ns_key=ns_key, kwargs=kwargs)


class DirParamType(Serializable):
    """The direction parameters container"""
    _fields = (
        'UVectECF', 'SS', 'ImpRespWid', 'Sgn', 'ImpRespBW', 'KCtr', 'DeltaK1', 'DeltaK2', 'DeltaKCOAPoly',
        'WgtType', 'WgtFunct')
    _required = ('UVectECF', 'SS', 'ImpRespWid', 'Sgn', 'ImpRespBW', 'KCtr', 'DeltaK1', 'DeltaK2')
    _numeric_format = {
        'SS': FLOAT_FORMAT, 'ImpRespWid': FLOAT_FORMAT, 'Sgn': '+d',
        'ImpRespBW': FLOAT_FORMAT, 'KCtr': FLOAT_FORMAT,
        'DeltaK1': FLOAT_FORMAT, 'DeltaK2': FLOAT_FORMAT, 'WgtFunct': FLOAT_FORMAT}
    _collections_tags = {'WgtFunct': {'array': True, 'child_tag': 'Wgt'}}
    # descriptors
    UVectECF = UnitVectorDescriptor(
        'UVectECF', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='Unit vector in the increasing ``(row/col)`` direction *(ECF)* at '
                  'the SCP pixel.')  # type: XYZType
    SS = FloatDescriptor(
        'SS', _required, strict=DEFAULT_STRICT,
        docstring='Sample spacing in the increasing ``(row/col)`` direction. Precise spacing '
                  'at the SCP.')  # type: float
    ImpRespWid = FloatDescriptor(
        'ImpRespWid', _required, strict=DEFAULT_STRICT,
        docstring='Half power impulse response width in the increasing ``(row/col)`` direction. '
                  'Measured at the scene center point.')  # type: float
    Sgn = IntegerEnumDescriptor(
        'Sgn', (1, -1), _required, strict=DEFAULT_STRICT,
        docstring='Sign for exponent in the DFT to transform the ``(row/col)`` dimension to '
                  'spatial frequency dimension.')  # type: int
    ImpRespBW = FloatDescriptor(
        'ImpRespBW', _required, strict=DEFAULT_STRICT,
        docstring='Spatial bandwidth in ``(row/col)`` used to form the impulse response in '
                  'the ``(row/col)`` direction. Measured at the center of '
                  'support for the SCP.')  # type: float
    KCtr = FloatDescriptor(
        'KCtr', _required, strict=DEFAULT_STRICT,
        docstring='Center spatial frequency in the given dimension. '
                  'Corresponds to the zero frequency of the DFT in the given ``(row/col)`` '
                  'direction.')  # type: float
    DeltaK1 = FloatDescriptor(
        'DeltaK1', _required, strict=DEFAULT_STRICT,
        docstring='Minimum ``(row/col)`` offset from KCtr of the spatial frequency support '
                  'for the image.')  # type: float
    DeltaK2 = FloatDescriptor(
        'DeltaK2', _required, strict=DEFAULT_STRICT,
        docstring='Maximum ``(row/col)`` offset from KCtr of the spatial frequency '
                  'support for the image.')  # type: float
    DeltaKCOAPoly = SerializableDescriptor(
        'DeltaKCOAPoly', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='Offset from KCtr of the center of support in the given ``(row/col)`` spatial frequency. '
                  'The polynomial is a function of image given ``(row/col)`` coordinate ``(variable 1)`` and '
                  'column coordinate ``(variable 2)``.')  # type: Poly2DType
    WgtType = SerializableDescriptor(
        'WgtType', WgtTypeType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters describing aperture weighting type applied in the spatial frequency domain '
                  'to yield the impulse response in the given ``(row/col)`` direction.')  # type: WgtTypeType
    WgtFunct = FloatArrayDescriptor(
        'WgtFunct', _collections_tags, _required, strict=DEFAULT_STRICT, minimum_length=2,
        docstring='Sampled aperture amplitude weighting function (array) applied to form the SCP impulse '
                  'response in the given ``(row/col)`` direction.')  # type: numpy.ndarray

    def __init__(
            self,
            UVectECF: Union[XYZType, numpy.ndarray, list, tuple] = None,
            SS: float = None,
            ImpRespWid: float = None,
            Sgn: int = None,
            ImpRespBW: float = None,
            KCtr: float = None,
            DeltaK1: float = None,
            DeltaK2: float = None,
            DeltaKCOAPoly: Union[None, Poly2DType, numpy.ndarray, list, tuple] = None,
            WgtType: Optional[WgtTypeType] = None,
            WgtFunct: Union[None, numpy.ndarray, list, tuple] = None,
            **kwargs):
        """

        Parameters
        ----------
        UVectECF : XYZType|numpy.ndarray|list|tuple
        SS : float
        ImpRespWid : float
        Sgn : int
        ImpRespBW : float
        KCtr : float
        DeltaK1 : float
        DeltaK2 : float
        DeltaKCOAPoly : Poly2DType|numpy.ndarray|list|tuple
        WgtType : WgtTypeType
        WgtFunct : None|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.UVectECF = UVectECF
        self.SS = SS
        self.ImpRespWid, self.ImpRespBW = ImpRespWid, ImpRespBW
        self.Sgn = Sgn
        self.KCtr, self.DeltaK1, self.DeltaK2 = KCtr, DeltaK1, DeltaK2
        self.DeltaKCOAPoly = DeltaKCOAPoly
        self.WgtType = WgtType
        self.WgtFunct = WgtFunct
        super(DirParamType, self).__init__(**kwargs)

    def define_weight_function(
            self,
            weight_size: int = DEFAULT_WEIGHT_SIZE,
            populate: bool = False) -> Optional[numpy.ndarray]:
        """
        Try to derive WgtFunct from WgtType, if necessary. This should likely be called from the `GridType` parent.

        Parameters
        ----------
        weight_size : int
            the size of the `WgtFunct` to generate.
        populate : bool
            Overwrite any populated WgtFunct value?

        Returns
        -------
        None|numpy.ndarray
        """

        if self.WgtType is None or self.WgtType.WindowName is None:
            return  # nothing to be done

        value = None
        window_name = self.WgtType.WindowName.upper()
        if window_name == 'HAMMING':
            # A Hamming window is defined in many places as a raised cosine of weight .54, so this is the default.
            # Some data use a generalized raised cosine and call it HAMMING, so we allow for both uses.
            try:
                # noinspection PyTypeChecker
                coef = float(self.WgtType.get_parameter_value(None, 0.54))  # just get first parameter - name?
            except ValueError:
                coef = 0.54
            value = general_hamming(weight_size, coef, sym=True)
        elif window_name == 'HANNING':
            value = general_hamming(weight_size, 0.5, sym=True)
        elif window_name == 'KAISER':
            beta = 14.0  # suggested default in literature/documentation
            try:
                # noinspection PyTypeChecker
                beta = float(self.WgtType.get_parameter_value(None, beta))  # just get first parameter - name?
            except ValueError:
                pass
            value = kaiser(weight_size, beta, sym=True)
        elif window_name == 'TAYLOR':
            # noinspection PyTypeChecker
            sidelobes = int(self.WgtType.get_parameter_value('NBAR', 4))  # apparently the matlab argument name
            # noinspection PyTypeChecker
            max_sidelobe_level = float(self.WgtType.get_parameter_value('SLL', -30))  # same
            value = taylor(weight_size, nbar=sidelobes, sll=max_sidelobe_level, norm=True, sym=True)
        elif window_name == 'UNIFORM':
            value = numpy.ones((32, ), dtype='float64')

        if self.WgtFunct is None or (populate and value is not None):
            self.WgtFunct = value
        return value

    def get_oversample_rate(self) -> float:
        """
        Gets the oversample rate. *Added in version 1.2.35.*

        Returns
        -------
        float
        """

        if self.SS is None or self.ImpRespBW is None:
            raise AttributeError('Both SS and ImpRespBW must be populated.')

        return max(1., 1./(self.SS*self.ImpRespBW))

    def _get_broadening_factor(self) -> float:
        """
        Gets the *broadening factor*, assuming that `WgtFunct` has been properly populated.

        Returns
        -------
        float
            the broadening factor
        """

        if self.WgtType is not None and self.WgtType.WindowName is not None:
            window_name = self.WgtType.WindowName.upper()
            coef = None
            if window_name == 'UNIFORM':
                coef = 1.0
            elif window_name == 'HAMMING':
                try:
                    # noinspection PyTypeChecker
                    coef = float(self.WgtType.get_parameter_value(None, 0.54))  # just get first parameter - name?
                except ValueError:
                    coef = 0.54
            elif window_name == 'HANNING':
                coef = 0.5

            if coef is not None:
                return get_hamming_broadening_factor(coef)

        return find_half_power(self.WgtFunct, oversample=1024)

    def define_response_widths(self, populate: bool = False) -> Optional[Tuple[float, float]]:
        """
        Assuming that `WgtFunct` has been properly populated, define the response widths.
        This should likely be called by `GridType` parent.

        Parameters
        ----------
        populate : bool
            Overwrite populated ImpRespWid and/or ImpRespBW?

        Returns
        -------
        None|(float, float)
            None or `(ImpRespBw, ImpRespWid)`
        """

        broadening_factor = self._get_broadening_factor()
        if broadening_factor is None:
            return None

        if self.ImpRespBW is not None:
            resp_width = broadening_factor/self.ImpRespBW
            if populate or self.ImpRespWid is None:
                self.ImpRespWid = resp_width
            return self.ImpRespBW, resp_width
        elif self.ImpRespWid is not None:
            resp_bw = broadening_factor/self.ImpRespWid
            if populate or self.ImpRespBW is None:
                self.ImpRespBW = resp_bw
            return resp_bw, self.ImpRespWid
        return None

    def estimate_deltak(
            self,
            x_coords: Optional[numpy.ndarray],
            y_coords: Optional[numpy.ndarray],
            populate: bool = False) -> Optional[Tuple[float, float]]:
        """
        The `DeltaK1` and `DeltaK2` parameters can be estimated from `DeltaKCOAPoly`, if necessary.
        This should likely be called by the `GridType` parent.

        Parameters
        ----------
        x_coords : None|numpy.ndarray
            The physical vertex coordinates to evaluate DeltaKCOAPoly
        y_coords : None|numpy.ndarray
            The physical vertex coordinates to evaluate DeltaKCOAPoly
        populate : bool
            Overwite any populated DeltaK1 and DeltaK2?

        Returns
        -------
        min_deltak: float
        max_deltak: float
        """

        if self.ImpRespBW is None or self.SS is None:
            return None  # nothing can be done

        if self.DeltaKCOAPoly is not None and x_coords is not None:
            deltaks = self.DeltaKCOAPoly(x_coords, y_coords)
            min_deltak = numpy.amin(deltaks) - 0.5*self.ImpRespBW
            max_deltak = numpy.amax(deltaks) + 0.5*self.ImpRespBW
        else:
            min_deltak = -0.5*self.ImpRespBW
            max_deltak = 0.5*self.ImpRespBW

        if (min_deltak < -0.5/abs(self.SS)) or (max_deltak > 0.5/abs(self.SS)):
            min_deltak = -0.5/abs(self.SS)
            max_deltak = -min_deltak

        if populate or (self.DeltaK1 is None or self.DeltaK2 is None):
            self.DeltaK1 = min_deltak
            self.DeltaK2 = max_deltak
        return min_deltak, max_deltak

    def check_deltak(
            self,
            x_coords: Optional[numpy.ndarray],
            y_coords: Optional[numpy.ndarray]) -> bool:
        """
        Checks the DeltaK values for validity.

        Parameters
        ----------
        x_coords : None|numpy.ndarray
            The physical vertex coordinates to evaluate DeltaKCOAPoly
        y_coords : None|numpy.ndarray
            The physical vertex coordinates to evaluate DeltaKCOAPoly

        Returns
        -------
        bool
        """

        out = True
        try:
            if self.DeltaK2 <= self.DeltaK1 + 1e-10:
                self.log_validity_error(
                    'DeltaK2 ({}) must be greater than DeltaK1 ({})'.format(self.DeltaK2, self.DeltaK1))
                out = False
        except (AttributeError, TypeError, ValueError):
            pass

        try:
            if self.DeltaK2 > 1./(2*self.SS) + 1e-10:
                self.log_validity_error(
                    'DeltaK2 ({}) must be <= 1/(2*SS) ({})'.format(self.DeltaK2, 1./(2*self.SS)))
                out = False
        except (AttributeError, TypeError, ValueError):
            pass

        try:
            if self.DeltaK1 < -1./(2*self.SS) - 1e-10:
                self.log_validity_error(
                    'DeltaK1 ({}) must be >= -1/(2*SS) ({})'.format(self.DeltaK1, -1./(2*self.SS)))
                out = False
        except (AttributeError, TypeError, ValueError):
            pass

        min_deltak, max_deltak = self.estimate_deltak(x_coords, y_coords, populate=False)
        try:
            if abs(self.DeltaK1/min_deltak - 1) > 1e-2:
                self.log_validity_error(
                    'The DeltaK1 value is populated as {}, but estimated to be {}'.format(self.DeltaK1, min_deltak))
                out = False
        except (AttributeError, TypeError, ValueError):
            pass

        try:
            if abs(self.DeltaK2/max_deltak - 1) > 1e-2:
                self.log_validity_error(
                    'The DeltaK2 value is populated as {}, but estimated to be {}'.format(self.DeltaK2, max_deltak))
                out = False
        except (AttributeError, TypeError, ValueError):
            pass
        return out

    def _check_bw(self) -> bool:
        out = True
        try:
            if self.ImpRespBW > (self.DeltaK2 - self.DeltaK1) + 1e-10:
                self.log_validity_error(
                    'ImpRespBW ({}) must be <= DeltaK2 - DeltaK1 '
                    '({})'.format(self.ImpRespBW, self.DeltaK2 - self.DeltaK1))
                out = False
        except (AttributeError, TypeError, ValueError):
            pass
        return out

    def _check_wgt(self) -> bool:
        cond = True
        if self.WgtType is None:
            return cond

        wgt_size = self.WgtFunct.size if self.WgtFunct is not None else None
        if self.WgtType.WindowName not in ['UNIFORM', 'UNKNOWN'] and (wgt_size is None or wgt_size < 2):
            self.log_validity_error(
                'Non-uniform weighting indicated, but WgtFunct not properly defined')
            return False

        if wgt_size is not None and wgt_size > 1024:
            self.log_validity_warning(
                'WgtFunct with {} elements is provided.\n'
                'The recommended number of elements is 512, '
                'and many more is likely needlessly excessive.'.format(wgt_size))

        result = self.define_response_widths(populate=False)
        if result is None:
            return cond
        resp_bw, resp_wid = result
        if abs(resp_bw/self.ImpRespBW - 1) > 1e-2:
            self.log_validity_error(
                'ImpRespBW expected as {} from weighting,\n'
                'but populated as {}'.format(resp_bw, self.ImpRespBW))
            cond = False
        if abs(resp_wid/self.ImpRespWid - 1) > 1e-2:
            self.log_validity_error(
                'ImpRespWid expected as {} from weighting,\n'
                'but populated as {}'.format(resp_wid, self.ImpRespWid))
            cond = False
        return cond

    def _basic_validity_check(self) -> bool:
        condition = super(DirParamType, self)._basic_validity_check()
        if (self.WgtFunct is not None) and (self.WgtFunct.size < 2):
            self.log_validity_error(
                'The WgtFunct array has been defined in DirParamType, '
                'but there are fewer than 2 entries.')
            condition = False
        for attribute in ['SS', 'ImpRespBW', 'ImpRespWid']:
            value = getattr(self, attribute)
            if value is not None and value <= 0:
                self.log_validity_error(
                    'attribute {} is populated as {}, '
                    'but should be strictly positive.'.format(attribute, value))
                condition = False
        condition &= self._check_bw()
        condition &= self._check_wgt()
        return condition


class GridType(Serializable):
    """
    Collection grid details container
    """

    _fields = ('ImagePlane', 'Type', 'TimeCOAPoly', 'Row', 'Col')
    _required = _fields
    _IMAGE_PLANE_VALUES = ('SLANT', 'GROUND', 'OTHER')
    _TYPE_VALUES = ('RGAZIM', 'RGZERO', 'XRGYCR', 'XCTYAT', 'PLANE')
    # descriptors
    ImagePlane = StringEnumDescriptor(
        'ImagePlane', _IMAGE_PLANE_VALUES, _required, strict=DEFAULT_STRICT,
        docstring="Defines the type of image plane that the best describes the sample grid. Precise plane "
                  "defined by Row Direction and Column Direction unit vectors.")  # type: str
    Type = StringEnumDescriptor(
        'Type', _TYPE_VALUES, _required, strict=DEFAULT_STRICT,
        docstring="""
        Defines the type of spatial sampling grid represented by the image sample grid. 
        Row coordinate first, column coordinate second:

        * `RGAZIM` - Grid for a simple range, Doppler image. Also, the natural grid for images formed with the Polar 
          Format Algorithm.

        * `RGZERO` - A grid for images formed with the Range Migration Algorithm. Used only for imaging near closest 
          approach (i.e. near zero Doppler).

        * `XRGYCR` - Orthogonal slant plane grid oriented range and cross range relative to the ARP at a 
          reference time.

        * `XCTYAT` - Orthogonal slant plane grid with X oriented cross track.

        * `PLANE` - Arbitrary plane with orientation other than the specific `XRGYCR` or `XCTYAT`.
        \n\n
        """)  # type: str
    TimeCOAPoly = SerializableDescriptor(
        'TimeCOAPoly', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring="*Time of Center Of Aperture* as a polynomial function of image coordinates. "
                  "The polynomial is a function of image row coordinate ``(variable 1)`` and column coordinate "
                  "``(variable 2)``.")  # type: Poly2DType
    Row = SerializableDescriptor(
        'Row', DirParamType, _required, strict=DEFAULT_STRICT,
        docstring="Row direction parameters.")  # type: DirParamType
    Col = SerializableDescriptor(
        'Col', DirParamType, _required, strict=DEFAULT_STRICT,
        docstring="Column direction parameters.")  # type: DirParamType

    def __init__(
            self,
            ImagePlane: str = None,
            Type: str = None,
            TimeCOAPoly: Union[Poly2DType, numpy.ndarray, list, tuple] = None,
            Row: DirParamType = None,
            Col: DirParamType = None,
            **kwargs):
        """

        Parameters
        ----------
        ImagePlane : str
        Type : str
        TimeCOAPoly : Poly2DType|numpy.ndarray|list|tuple
        Row : DirParamType
        Col : DirParamType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ImagePlane = ImagePlane
        self.Type = Type
        self.TimeCOAPoly = TimeCOAPoly
        self.Row, self.Col = Row, Col
        super(GridType, self).__init__(**kwargs)

    def derive_direction_params(
            self,
            ImageData,
            populate: bool = False):
        """
        Populate the ``Row/Col`` direction parameters from ImageData, if necessary.
        Expected to be called from SICD parent.

        Parameters
        ----------
        ImageData : sarpy.io.complex.sicd_elements.ImageData.ImageDataType
        populate : bool
            Repopulates any present values?

        Returns
        -------
        None
        """

        valid_vertices = None
        if ImageData is not None:
            valid_vertices = ImageData.get_valid_vertex_data()
            if valid_vertices is None:
                valid_vertices = ImageData.get_full_vertex_data()

        x_coords, y_coords = None, None
        if valid_vertices is not None:
            try:
                x_coords = self.Row.SS*(valid_vertices[:, 0] - (ImageData.SCPPixel.Row - ImageData.FirstRow))
                y_coords = self.Col.SS*(valid_vertices[:, 1] - (ImageData.SCPPixel.Col - ImageData.FirstCol))
            except (AttributeError, ValueError):
                pass

        for attribute in ['Row', 'Col']:
            value = getattr(self, attribute, None)
            if value is not None:
                value.define_weight_function(populate=populate)
                value.define_response_widths(populate=populate)
                value.estimate_deltak(x_coords, y_coords, populate=populate)

    def _derive_time_coa_poly(
            self,
            CollectionInfo,
            SCPCOA):
        """
        Expected to be called from SICD parent.

        Parameters
        ----------
        CollectionInfo : sarpy.io.complex.sicd_elements.CollectionInfo.CollectionInfoType
        SCPCOA : sarpy.io.complex.sicd_elements.SCPCOA.SCPCOAType

        Returns
        -------
        None
        """
        if self.TimeCOAPoly is not None:
            return  # nothing needs to be done

        try:
            if CollectionInfo.RadarMode.ModeType == 'SPOTLIGHT':
                self.TimeCOAPoly = Poly2DType(Coefs=[[SCPCOA.SCPTime, ], ])
        except (AttributeError, ValueError):
            return

    def _derive_rg_az_comp(
            self,
            GeoData,
            SCPCOA,
            RadarCollection,
            ImageFormation):
        """
        Expected to be called by SICD parent.

        Parameters
        ----------
        GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType
        SCPCOA : sarpy.io.complex.sicd_elements.SCPCOA.SCPCOAType
        RadarCollection : sarpy.io.complex.sicd_elements.RadarCollection.RadarCollectionType
        ImageFormation : sarpy.io.complex.sicd_elements.ImageFormation.ImageFormationType

        Returns
        -------
        None
        """

        if self.Row is None:
            self.Row = DirParamType()
        if self.Col is None:
            self.Col = DirParamType()

        if self.ImagePlane is None:
            self.ImagePlane = 'SLANT'
        elif self.ImagePlane != 'SLANT':
            logger.warning(
                'The Grid.ImagePlane is set to {},\n\t'
                'but Image Formation Algorithm is RgAzComp, which requires "SLANT".\n\t'
                'Resetting.'.format(self.ImagePlane))
            self.ImagePlane = 'SLANT'

        if self.Type is None:
            self.Type = 'RGAZIM'
        elif self.Type != 'RGAZIM':
            logger.warning(
                'The Grid.Type is set to {},\n\t'
                'but Image Formation Algorithm is RgAzComp, which requires "RGAZIM".\n\t'
                'Resetting.'.format(self.Type))
            self.Type = 'RGAZIM'

        if GeoData is not None and GeoData.SCP is not None and GeoData.SCP.ECF is not None and \
                SCPCOA.ARPPos is not None and SCPCOA.ARPVel is not None:
            scp = GeoData.SCP.ECF.get_array()
            arp = SCPCOA.ARPPos.get_array()
            los = (scp - arp)
            ulos = los/norm(los)
            if self.Row.UVectECF is None:
                self.Row.UVectECF = XYZType.from_array(ulos)

            look = SCPCOA.look
            arp_vel = SCPCOA.ARPVel.get_array()
            uspz = look*numpy.cross(arp_vel, ulos)
            uspz /= norm(uspz)
            uaz = numpy.cross(uspz, ulos)
            if self.Col.UVectECF is None:
                self.Col.UVectECF = XYZType.from_array(uaz)

        center_frequency = _get_center_frequency(RadarCollection, ImageFormation)
        if center_frequency is not None:
            if self.Row.KCtr is None:
                kctr = 2*center_frequency/speed_of_light
                if self.Row.DeltaKCOAPoly is not None:  # assume it's 0 otherwise?
                    kctr -= self.Row.DeltaKCOAPoly.Coefs[0, 0]
                self.Row.KCtr = kctr
            elif self.Row.DeltaKCOAPoly is None:
                self.Row.DeltaKCOAPoly = Poly2DType(Coefs=[[2*center_frequency/speed_of_light - self.Row.KCtr, ], ])

            if self.Col.KCtr is None:
                kctr = 0.0
                if self.Col.DeltaKCOAPoly is not None:
                    kctr -= self.Col.DeltaKCOAPoly.Coefs[0, 0]
                self.Col.KCtr = kctr
            elif self.Col.DeltaKCOAPoly is None:
                self.Col.DeltaKCOAPoly = Poly2DType(Coefs=[[-self.Col.KCtr, ], ])

    def _derive_pfa(
            self,
            GeoData,
            RadarCollection,
            ImageFormation,
            Position,
            PFA):
        """
        Expected to be called by SICD parent.

        Parameters
        ----------
        GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType
        RadarCollection : sarpy.io.complex.sicd_elements.RadarCollection.RadarCollectionType
        ImageFormation : sarpy.io.complex.sicd_elements.ImageFormation.ImageFormationType
        Position : sarpy.io.complex.sicd_elements.Position.PositionType
        PFA : sarpy.io.complex.sicd_elements.PFA.PFAType

        Returns
        -------
        None
        """

        if self.Type is None:
            self.Type = 'RGAZIM'  # the natural result for PFA

        if PFA is None:
            return  # nothing to be done
        if GeoData is None or GeoData.SCP is None:
            return  # nothing to be done

        scp = GeoData.SCP.ECF.get_array()

        if Position is not None and Position.ARPPoly is not None \
                and PFA.PolarAngRefTime is not None:
            polar_ref_pos = Position.ARPPoly(PFA.PolarAngRefTime)
        else:
            polar_ref_pos = scp

        if PFA.IPN is not None and PFA.FPN is not None and \
                self.Row.UVectECF is None and self.Col.UVectECF is None:
            ipn = PFA.IPN.get_array()
            fpn = PFA.FPN.get_array()

            dist = numpy.dot((scp - polar_ref_pos), ipn) / numpy.dot(fpn, ipn)
            ref_pos_ipn = polar_ref_pos + (dist * fpn)
            urg = scp - ref_pos_ipn
            urg /= norm(urg)
            uaz = numpy.cross(ipn, urg)  # already unit
            self.Row.UVectECF = XYZType.from_array(urg)
            self.Col.UVectECF = XYZType.from_array(uaz)

        if self.Col is not None and self.Col.KCtr is None:
            self.Col.KCtr = 0  # almost always 0 for PFA

        if self.Row is not None and self.Row.KCtr is None:
            center_frequency = _get_center_frequency(RadarCollection, ImageFormation)
            if PFA.Krg1 is not None and PFA.Krg2 is not None:
                self.Row.KCtr = 0.5*(PFA.Krg1 + PFA.Krg2)
            elif center_frequency is not None and PFA.SpatialFreqSFPoly is not None:
                # APPROXIMATION: may not be quite right, due to rectangular inscription loss in PFA.
                self.Row.KCtr = 2*center_frequency/speed_of_light + PFA.SpatialFreqSFPoly.Coefs[0]

    def _derive_rma(
            self,
            RMA,
            GeoData,
            RadarCollection,
            ImageFormation,
            Position):
        """

        Parameters
        ----------
        RMA : sarpy.io.complex.sicd_elements.RMA.RMAType
        GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType
        RadarCollection : sarpy.io.complex.sicd_elements.RadarCollection.RadarCollectionType
        ImageFormation : sarpy.io.complex.sicd_elements.ImageFormation.ImageFormationType
        Position : sarpy.io.complex.sicd_elements.Position.PositionType

        Returns
        -------
        None
        """

        if RMA is None:
            return   # nothing can be derived

        im_type = RMA.ImageType

        if im_type is None:
            return
        if im_type == 'INCA':
            self._derive_rma_inca(RMA, GeoData, Position)
        else:
            if im_type == 'RMAT':
                self._derive_rma_rmat(RMA, GeoData, RadarCollection, ImageFormation)
            elif im_type == 'RMCR':
                self._derive_rma_rmcr(RMA, GeoData, RadarCollection, ImageFormation)

    @staticmethod
    def _derive_unit_vector_params(GeoData, RMAParam):
        """
        Helper method.

        Parameters
        ----------
        GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType
        RMAParam : sarpy.io.complex.sicd_elements.RMA.RMATType|sarpy.io.complex.sicd_elements.RMA.RMCRType

        Returns
        -------
        Tuple[numpy.ndarray,...]
        """

        if GeoData is None or GeoData.SCP is None:
            return None

        scp = GeoData.SCP.ECF.get_array()
        pos_ref = RMAParam.PosRef.get_array()
        upos_ref = pos_ref / norm(pos_ref)
        vel_ref = RMAParam.VelRef.get_array()
        uvel_ref = vel_ref / norm(vel_ref)
        los = (scp - pos_ref)  # it absolutely could be that scp = pos_ref
        los_norm = norm(los)
        if los_norm < 1:
            logger.error(
                msg="Row/Col UVectECF cannot be derived from RMA,\n\t"
                    "because the Reference Position is too close (less than 1 meter) to the SCP.")
        ulos = los/los_norm
        left = numpy.cross(upos_ref, uvel_ref)
        look = numpy.sign(numpy.dot(left, ulos))
        return scp, upos_ref, uvel_ref, ulos, left, look

    def _derive_rma_rmat(
            self,
            RMA,
            GeoData,
            RadarCollection,
            ImageFormation):
        """

        Parameters
        ----------
        RMA : sarpy.io.complex.sicd_elements.RMA.RMAType
        GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType
        RadarCollection : sarpy.io.complex.sicd_elements.RadarCollection.RadarCollectionType
        ImageFormation : sarpy.io.complex.sicd_elements.ImageFormation.ImageFormationType

        Returns
        -------
        None
        """

        if RMA.RMAT is None:
            return

        if self.ImagePlane is None:
            self.ImagePlane = 'SLANT'
        if self.Type is None:
            self.Type = 'XCTYAT'

        if self.Row.UVectECF is None and self.Col.UVectECF is None:
            params = self._derive_unit_vector_params(GeoData, RMA.RMAT)
            if params is not None:
                scp, upos_ref, uvel_ref, ulos, left, look = params
                uyat = -look*uvel_ref
                uspz = numpy.cross(ulos, uyat)
                uspz /= norm(uspz)
                uxct = numpy.cross(uyat, uspz)
                self.Row.UVectECF = XYZType.from_array(uxct)
                self.Col.UVectECF = XYZType.from_array(uyat)

        center_frequency = _get_center_frequency(RadarCollection, ImageFormation)
        if center_frequency is not None and RMA.RMAT.DopConeAngRef is not None:
            if self.Row.KCtr is None:
                self.Row.KCtr = (2*center_frequency/speed_of_light)*numpy.sin(numpy.deg2rad(RMA.RMAT.DopConeAngRef))
            if self.Col.KCtr is None:
                self.Col.KCtr = (2*center_frequency/speed_of_light)*numpy.cos(numpy.deg2rad(RMA.RMAT.DopConeAngRef))

    def _derive_rma_rmcr(
            self,
            RMA,
            GeoData,
            RadarCollection,
            ImageFormation):
        """

        Parameters
        ----------
        RMA : sarpy.io.complex.sicd_elements.RMA.RMAType
        GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType
        RadarCollection : sarpy.io.complex.sicd_elements.RadarCollection.RadarCollectionType
        ImageFormation : sarpy.io.complex.sicd_elements.ImageFormation.ImageFormationType

        Returns
        -------
        None
        """

        if RMA.RMCR is None:
            return

        if self.ImagePlane is None:
            self.ImagePlane = 'SLANT'
        if self.Type is None:
            self.Type = 'XRGYCR'

        if self.Row.UVectECF is None and self.Col.UVectECF is None:
            params = self._derive_unit_vector_params(GeoData, RMA.RMCR)
            if params is not None:
                scp, upos_ref, uvel_ref, ulos, left, look = params
                uxrg = ulos
                uspz = look*numpy.cross(uvel_ref, uxrg)
                uspz /= norm(uspz)
                uycr = numpy.cross(uspz, uxrg)
                self.Row.UVectECF = XYZType.from_array(uxrg)
                self.Col.UVectECF = XYZType.from_array(uycr)

        center_frequency = _get_center_frequency(RadarCollection, ImageFormation)
        if center_frequency is not None:
            if self.Row.KCtr is None:
                self.Row.KCtr = 2*center_frequency/speed_of_light
            if self.Col.KCtr is None:
                self.Col.KCtr = 2*center_frequency/speed_of_light

    def _derive_rma_inca(
            self,
            RMA,
            GeoData,
            Position):
        """

        Parameters
        ----------
        RMA : sarpy.io.complex.sicd_elements.RMA.RMAType
        GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType
        Position : sarpy.io.complex.sicd_elements.Position.PositionType

        Returns
        -------
        None
        """

        if RMA.INCA is None:
            return

        if self.Type is None:
            self.Type = 'RGZERO'

        if RMA.INCA.TimeCAPoly is not None and Position is not None and Position.ARPPoly is not None and \
                self.Row.UVectECF is None and self.Col.UVectECF is None and \
                GeoData is not None and GeoData.SCP is not None:
            scp = GeoData.SCP.ECF.get_array()

            t_zero = RMA.INCA.TimeCAPoly.Coefs[0]
            ca_pos = Position.ARPPoly(t_zero)
            ca_vel = Position.ARPPoly.derivative_eval(t_zero, der_order=1)

            uca_pos = ca_pos/norm(ca_pos)
            uca_vel = ca_vel/norm(ca_vel)
            urg = (scp - ca_pos)
            urg_norm = norm(urg)
            if urg_norm > 0:
                urg /= urg_norm
                left = numpy.cross(uca_pos, uca_vel)
                look = numpy.sign(numpy.dot(left, urg))
                uspz = -look*numpy.cross(urg, uca_vel)
                uspz /= norm(uspz)
                uaz = numpy.cross(uspz, urg)
                self.Row.UVectECF = XYZType.from_array(urg)
                self.Col.UVectECF = XYZType.from_array(uaz)

        if self.Row is not None and self.Row.KCtr is None and RMA.INCA.FreqZero is not None:
            self.Row.KCtr = 2*RMA.INCA.FreqZero/speed_of_light
        if self.Col is not None and self.Col.KCtr is None:
            self.Col.KCtr = 0

    def _basic_validity_check(self) -> bool:
        condition = super(GridType, self)._basic_validity_check()
        if self.Row is not None and self.Row.Sgn is not None and self.Col is not None \
                and self.Col.Sgn is not None and self.Row.Sgn != self.Col.Sgn:
            self.log_validity_warning(
                'Row.Sgn ({}) and Col.Sgn ({}) should almost certainly be the '
                'same value'.format(self.Row.Sgn, self.Col.Sgn))
        return condition

    def check_deltak(
            self,
            x_coords: Optional[numpy.ndarray],
            y_coords: Optional[numpy.ndarray]) -> bool:
        """
        Checks the validity of DeltaK values.

        Parameters
        ----------
        x_coords : None|numpy.ndarray
        y_coords : None|numpy.ndarray

        Returns
        -------
        bool
        """

        cond = True
        if self.Row is not None:
            cond &= self.Row.check_deltak(x_coords, y_coords)
        if self.Col is not None:
            cond &= self.Col.check_deltak(x_coords, y_coords)
        return cond

    def get_resolution_abbreviation(self) -> str:
        """
        Gets the resolution abbreviation for the suggested name.

        Returns
        -------
        str
        """

        if self.Row is None or self.Row.ImpRespWid is None or \
                self.Col is None or self.Col.ImpRespWid is None:
            return '0000'
        else:
            value = int(100*(abs(self.Row.ImpRespWid)*abs(self.Col.ImpRespWid))**0.5)
            if value > 9999:
                return '9999'
            else:
                return '{0:04d}'.format(value)

    def get_slant_plane_area(self) -> float:
        """
        Get the weighted slant plane area.

        Returns
        -------
        float
        """

        range_weight_f = azimuth_weight_f = 1.0
        if self.Row.WgtFunct is not None:
            var = numpy.var(self.Row.WgtFunct)
            mean = numpy.mean(self.Row.WgtFunct)
            range_weight_f += var/(mean*mean)
        if self.Col.WgtFunct is not None:
            var = numpy.var(self.Col.WgtFunct)
            mean = numpy.mean(self.Col.WgtFunct)
            azimuth_weight_f += var/(mean*mean)
        return (range_weight_f * azimuth_weight_f)/(self.Row.ImpRespBW*self.Col.ImpRespBW)



"""
The RgAzCompType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from typing import Union

import numpy
from numpy.linalg import norm

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import FloatDescriptor, SerializableDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import Poly1DType


logger = logging.getLogger(__name__)


class RgAzCompType(Serializable):
    """
    Parameters included for a Range, Doppler image.
    """

    _fields = ('AzSF', 'KazPoly')
    _required = _fields
    _numeric_format = {'AzSF': FLOAT_FORMAT}
    # descriptors
    AzSF = FloatDescriptor(
        'AzSF', _required, strict=DEFAULT_STRICT,
        docstring='Scale factor that scales image coordinate az = ycol (meters) to a delta cosine of the '
                  'Doppler Cone Angle at COA, *(in 1/m)*')  # type: float
    KazPoly = SerializableDescriptor(
        'KazPoly', Poly1DType, _required, strict=DEFAULT_STRICT,
        docstring='Polynomial function that yields azimuth spatial frequency *(Kaz = Kcol)* as a function of '
                  'slow time ``(variable 1)``. That is '
                  r':math:`\text{Slow Time (sec)} \to \text{Azimuth spatial frequency (cycles/meter)}`. '
                  'Time relative to collection start.')  # type: Poly1DType

    def __init__(
            self,
            AzSF: float = None,
            KazPoly: Union[Poly1DType, numpy.ndarray, list, tuple] = None,
            **kwargs):
        """

        Parameters
        ----------
        AzSF : float
        KazPoly : Poly1DType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.AzSF = AzSF
        self.KazPoly = KazPoly
        super(RgAzCompType, self).__init__(**kwargs)

    def _derive_parameters(self, Grid, Timeline, SCPCOA):
        """
        Expected to be called by the SICD object.

        Parameters
        ----------
        Grid : sarpy.io.complex.sicd_elements.GridType
        Timeline : sarpy.io.complex.sicd_elements.TimelineType
        SCPCOA : sarpy.io.complex.sicd_elements.SCPCOA.SCPCOAType

        Returns
        -------
        None
        """

        look = SCPCOA.look
        az_sf = -look*numpy.sin(numpy.deg2rad(SCPCOA.DopplerConeAng))/SCPCOA.SlantRange
        if self.AzSF is None:
            self.AzSF = az_sf
        elif abs(self.AzSF - az_sf) > 1e-3:
            logger.warning(
                'The derived value for RgAzComp.AzSF is {},\n\t'
                'while the current setting is {}.'.format(az_sf, self.AzSF))

        if self.KazPoly is None:
            if Grid.Row.KCtr is not None and Timeline is not None and Timeline.IPP is not None and \
                    Timeline.IPP.size == 1 and Timeline.IPP[0].IPPPoly is not None and SCPCOA.SCPTime is not None:

                st_rate_coa = Timeline.IPP[0].IPPPoly.derivative_eval(SCPCOA.SCPTime, 1)

                krg_coa = Grid.Row.KCtr
                if Grid.Row is not None and Grid.Row.DeltaKCOAPoly is not None:
                    krg_coa += Grid.Row.DeltaKCOAPoly.Coefs[0, 0]

                # Scale factor described in SICD spec
                delta_kaz_per_delta_v = \
                    look*krg_coa*norm(SCPCOA.ARPVel.get_array()) * \
                    numpy.sin(numpy.deg2rad(SCPCOA.DopplerConeAng))/(SCPCOA.SlantRange*st_rate_coa)
                self.KazPoly = Poly1DType(Coefs=delta_kaz_per_delta_v*Timeline.IPP[0].IPPPoly.Coefs)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PLTFMAType(TREElement):
    def __init__(self, value):
        super(PLTFMAType, self).__init__()
        self.add_field('VERNUM', 's', 4, value)
        self.add_field('P_NAME', 's', 12, value)
        self.add_field('P_DESCR', 's', 40, value)
        self.add_field('P_DATE', 's', 8, value)
        self.add_field('P_TIME', 's', 9, value)
        self.add_field('P_TYPE', 's', 1, value)
        if self.P_TYPE in ['G', 'T', 'M']:
            self.add_field('SNSR_HT', 's', 3, value)
            self.add_field('SNSRLOC', 's', 21, value)
            self.add_field('SNSRHDNG', 's', 3, value)
        elif self.P_TYPE == 'A':
            self.add_field('AC_TYPE', 's', 15, value)
            self.add_field('AC_SERIAL', 's', 12, value)
            self.add_field('AC_T_NUM', 's', 10, value)
            self.add_field('AC_PITCH', 's', 5, value)
            self.add_field('AC_ROLL', 's', 5, value)
            self.add_field('AC_HDNG', 's', 3, value)
            self.add_field('AC_REF_PT', 's', 1, value)
            self.add_field('AC_POS_X', 's', 9, value)
            self.add_field('AC_POS_Y', 's', 9, value)
            self.add_field('AC_POS_Z', 's', 9, value)
            self.add_field('AC_VEL_X', 's', 9, value)
            self.add_field('AC_VEL_Y', 's', 9, value)
            self.add_field('AC_POS_Z', 's', 9, value)
            self.add_field('AC_ACC_X', 's', 8, value)
            self.add_field('AC_ACC_Y', 's', 8, value)
            self.add_field('AC_POS_Z', 's', 8, value)
            self.add_field('AC_SPEED', 's', 5, value)
            self.add_field('ENTLOC', 's', 21, value)
            self.add_field('ENTALT', 's', 6, value)
            self.add_field('EXITLOC', 's', 21, value)
            self.add_field('EXITALTH', 's', 6, value)
            self.add_field('INS_V_NC', 's', 5, value)
            self.add_field('INS_V_EC', 's', 5, value)
            self.add_field('INS_V_DC', 's', 5, value)


class PLTFMA(TREExtension):
    _tag_value = 'PLTFMA'
    _data_type = PLTFMAType



"""
Functionality for reading SIO data into a SICD model.

The SIO format is believed to be based on a memo from General Dynamics.  Some SIO features may not
be implemented and compatibility with other SIO software is not guaranteed.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Wade Schwartzkopf")


import os
import struct
import logging
import re
from typing import Union, Dict, Tuple, Optional, BinaryIO

import numpy

from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.blocks import RowColType
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.ImageData import ImageDataType, FullImageType
from sarpy.io.complex.sicd import AmpLookupFunction

from sarpy.io.general.base import BaseWriter, SarpyIOError
from sarpy.io.general.data_segment import NumpyArraySegment, NumpyMemmapSegment
from sarpy.io.general.format_function import ComplexFormatFunction
from sarpy.io.general.utils import is_file_like, is_real_file
from sarpy.io.xml.base import parse_xml_from_string

logger = logging.getLogger(__name__)

_unsupported_pix_size = 'Got unsupported sio data type/pixel size = `{}`'


###########
# parser and interpreter for hdf5 attributes

class SIODetails(object):
    __slots__ = (
        '_file_name', '_magic_number', '_head', '_user_data', '_data_offset',
        '_caspr_data', '_reverse_axes', '_transpose_axes', '_sicd')

    # NB: there are really just two types of SIO file (with user_data and without),
    #   with endian-ness layered on top
    ENDIAN = {
        0xFF017FFE: '>', 0xFE7F01FF: '<',  # no user data
        0xFF027FFD: '>', 0xFD7F02FF: '<'}  # with user data

    def __init__(self, file_name: str):
        self._file_name = file_name
        self._user_data = None
        self._data_offset = 20
        self._caspr_data = None
        self._reverse_axes = None
        self._transpose_axes = None
        self._sicd = None

        if not os.path.isfile(file_name):
            raise SarpyIOError('Path {} is not a file'.format(file_name))

        with open(file_name, 'rb') as fi:
            self._magic_number = struct.unpack(">I", fi.read(4))[0]
            endian = self.ENDIAN.get(self._magic_number, None)
            if endian is None:
                raise SarpyIOError(
                    'File {} is not an SIO file. Got magic number {}'.format(file_name, self._magic_number))

            # reader basic header - (rows, columns, data_type, pixel_size)?
            init_head = numpy.array(struct.unpack('{}4I'.format(endian), fi.read(16)), dtype=numpy.uint64)
            if not (numpy.all(init_head[2:] == numpy.array([13, 8]))
                    or numpy.all(init_head[2:] == numpy.array([12, 4]))
                    or numpy.all(init_head[2:] == numpy.array([11, 2]))):
                raise SarpyIOError(_unsupported_pix_size.format(init_head[2:]))
            self._head = init_head

    @property
    def file_name(self) -> str:
        return self._file_name

    @property
    def data_offset(self):  # type: () -> int
        return self._data_offset

    @property
    def raw_data_size(self) -> Optional[Tuple[int, ...]]:
        if self._head is None:
            return None
        rows, cols = self._head[:2]
        return int(rows), int(cols), 2

    @property
    def formatted_data_size(self) -> Union[None, Tuple[int, ...]]:
        if self._head is None:
            return None
        rows, cols = self._head[:2]
        if self._transpose_axes is not None:
            return int(cols), int(rows)
        else:
            return int(rows), int(cols)

    @property
    def raw_data_type(self):  # type: () -> Union[None, str]
        # head[2] = (2X = vector, 1X = complex/scalar, 0X = real/scalar), where
        #   X = (1 = unsigned int, 2 = signed int, 3 = float, (4=double? I would guess)
        # head[3] = pixel size in bytes (2*bit depth for complex, or band*bit depth for vector)
        pixel_size = self._head[3]
        endian = self.ENDIAN[self._magic_number]
        # we require (for sicd) that either head[2:] is [13, 8], [12, 4], [11, 2]
        if pixel_size == 8:
            return '{}f4'.format(endian)
        elif pixel_size == 4:
            return '{}i2'.format(endian)
        elif pixel_size == 2:
            return '{}u1'.format(endian)
        else:
            raise ValueError(_unsupported_pix_size.format(self._head[2:]))

    @property
    def pixel_type(self) -> str:
        if self._head[2] == 13 and self._head[3] == 8:
            return 'RE32F_IM32F'
        elif self._head[2] == 12 and self._head[3] == 4:
            return 'RE16I_IM16I'
        elif self._head[2] == 11 and self._head[3] == 2:
            return 'AMP8I_PHS8I'
        else:
            raise ValueError(_unsupported_pix_size.format(self._head[2:]))

    def get_symmetry(self) -> Tuple[Optional[Tuple[int, ...]], Optional[Tuple[int, ...]]]:
        return self._reverse_axes, self._transpose_axes

    def _read_user_data(self):
        if self._user_data is not None:
            return
        if self._magic_number in (0xFF017FFE, 0xFE7F01FF):  # no user data
            self._user_data = {}
        else:
            def read_user_data():
                out = {}
                user_dat_len = 0
                if self._magic_number in (0xFF017FFE, 0xFE7F01FF):  # no user data
                    return out, user_dat_len

                num_data_pairs = struct.unpack('{}I'.format(endian), fi.read(4))[0]
                user_dat_len = 4

                for i in range(num_data_pairs):
                    name_length = struct.unpack('{}I'.format(endian), fi.read(4))[0]
                    name = struct.unpack('{}{}s'.format(endian, name_length), fi.read(name_length))[0].decode('utf-8')
                    value_length = struct.unpack('{}I'.format(endian), fi.read(4))[0]
                    value = struct.unpack('{}{}s'.format(endian, value_length), fi.read(value_length))[0]
                    try:
                        value = value.decode('utf-8')
                    except UnicodeDecodeError:
                        # leave value as bytes - it may just be some other type
                        pass
                    out[name] = value
                    user_dat_len += 4 + name_length + 4 + value_length
                return out, user_dat_len

            endian = self.ENDIAN[self._magic_number]
            with open(self._file_name, 'rb') as fi:
                fi.seek(20, os.SEEK_SET)  # skip the basic header
                # read the user data (some type of header), if necessary
                user_data, user_data_length = read_user_data()
            self._user_data = user_data
            self._data_offset = 20 + user_data_length
        # validate file size
        exp_file_size = self._data_offset + self._head[0]*self._head[1]*self._head[3]
        act_file_size = os.path.getsize(self._file_name)
        if exp_file_size != act_file_size:
            logger.warning(
                'File {} appears to be an SIO file.\n\t'
                'The file size calculated from the header ({})\n\t'
                'does not match the actual file size ({})'.format(
                    self._file_name, exp_file_size, act_file_size))

    def _find_caspr_data(self) -> None:
        def find_caspr():
            dir_name, fil_name = os.path.split(self._file_name)
            file_stem = os.path.splitext(fil_name)
            for fil in [os.path.join(dir_name, '{}.hydra'.format(file_stem)),
                        os.path.join(dir_name, '{}.hdr'.format(file_stem)),
                        os.path.join(dir_name, '..', 'RPHDHeader.out'),
                        os.path.join(dir_name, '..', '_RPHDHeader.out')]:
                if os.path.exists(fil) and os.path.isfile(fil):
                    # generally redundant, except for broken link?
                    return fil
            return None

        casp_fil = find_caspr()
        if casp_fil is None:
            return

        out = {}
        with open(casp_fil, 'r') as fi:
            lines = fi.read().splitlines(keepends=False)
        # this is generally just copied from the previous version - maybe refactor eventually
        current_subfield = ''
        reading_subfield = False
        for line in lines:
            if len(line.strip()) == 0:
                continue  # skip blank lines
            if line.startswith(';;;'):  # I guess this is the subfield delimiter?
                reading_subfield = ~reading_subfield  # change state
                if reading_subfield:
                    current_subfield = ''
                else:
                    out[current_subfield] = {}  # prepare the workspace
            else:
                quoted_token = re.match('"(?P<quoted>[^"]+)"', line)
                if quoted_token:  # Some values with spaces are surrounded by quotes
                    tokens = [quoted_token.group('quoted'),
                              line[quoted_token.end('quoted') + 1:].strip()]
                else:  # No quoted values were found
                    # If not using quotes, split with whitespace
                    tokens = line.split(None, 1)
                if (len(tokens) > 1) and tokens[1] != '':
                    if reading_subfield:  # Subsection heading
                        current_subfield = current_subfield + tokens[1]
                    elif not current_subfield == '':  # Actual field value
                        try:
                            out[current_subfield][tokens[1]] = float(tokens[0])
                        except ValueError:  # Value is string not numeric
                            out[current_subfield][tokens[1]] = tokens[0]
        self._caspr_data = out
        # set symmetry
        im_params = out.get('Image Parameters', None)
        if im_params is None:
            return
        illum_dir = im_params.get('image illumination direction [top, left, bottom, right]', None)
        if illum_dir is None:
            return
        elif illum_dir == 'left':
            self._reverse_axes = (0, )
            self._transpose_axes = (1, 0, 2)
        elif illum_dir != 'top':
            raise ValueError('unhandled illumination direction {}'.format(illum_dir))

    def get_sicd(self) -> SICDType:
        """
        Extract the SICD details.

        Returns
        -------
        SICDType
        """

        if self._sicd is not None:
            return self._sicd
        if self._user_data is None:
            self._read_user_data()
        if self._caspr_data is None:
            self._find_caspr_data()
        # Check if the user data contains a sicd structure.
        sicd_string = None
        for nam in ['SICDMETA', 'SICD_META', 'SICD']:
            if sicd_string is None:
                sicd_string = self._user_data.get(nam, None)
        # If so, assume that this SICD is valid and simply present it
        if sicd_string is not None:
            root_node, xml_ns = parse_xml_from_string(sicd_string)
            self._sicd = SICDType.from_node(root_node, xml_ns)
            self._sicd.derive()
        else:
            # otherwise, we populate a really minimal sicd structure
            num_rows, num_cols = self.formatted_data_size
            self._sicd = SICDType(ImageData=ImageDataType(NumRows=num_rows,
                                                          NumCols=num_cols,
                                                          FirstRow=0,
                                                          FirstCol=0,
                                                          PixelType=self.pixel_type,
                                                          FullImage=FullImageType(NumRows=num_rows,
                                                                                  NumCols=num_cols),
                                                          SCPPixel=RowColType(Row=num_rows/2,
                                                                              Col=num_cols/2)))
        return self._sicd


#######
#  The actual reading implementation

class SIOReader(SICDTypeReader):
    """
    **Changed in version 1.3.0** for reading changes.
    """
    __slots__ = ('_sio_details', )

    def __init__(self, sio_details):
        """

        Parameters
        ----------
        sio_details : str|SIODetails
            filename or SIODetails object
        """

        if isinstance(sio_details, str):
            sio_details = SIODetails(sio_details)
        if not isinstance(sio_details, SIODetails):
            raise TypeError('The input argument for SIOReader must be a filename or '
                            'SIODetails object.')
        self._sio_details = sio_details
        sicd_meta = sio_details.get_sicd()

        if sicd_meta.ImageData.PixelType == 'AMP8I_PHS8I':
            format_function = AmpLookupFunction(sio_details.raw_data_type, sicd_meta.ImageData.AmpTable)
        else:
            format_function = ComplexFormatFunction(
                sio_details.raw_data_type, order='IQ', band_dimension=-1)
        reverse_axes, transpose_axes = sio_details.get_symmetry()
        data_segment = NumpyMemmapSegment(
            sio_details.file_name, sio_details.data_offset,
            sio_details.raw_data_type, sio_details.raw_data_size,
            'complex64', sio_details.formatted_data_size,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes,
            format_function=format_function, mode='r', close_file=True)

        SICDTypeReader.__init__(self, data_segment, sicd_meta, close_segments=True)
        self._check_sizes()

    @property
    def sio_details(self) -> SIODetails:
        """
        SIODetails: The sio details object.
        """

        return self._sio_details

    @property
    def file_name(self) -> str:
        return self.sio_details.file_name


########
# base expected functionality for a module with an implemented Reader

def is_a(file_name: str) -> Optional[SIOReader]:
    """
    Tests whether a given file_name corresponds to a SIO file. Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str
        the file_name to check

    Returns
    -------
    SIOReader|None
        `SIOReader` instance if SIO file, `None` otherwise
    """

    if is_file_like(file_name):
        return None

    try:
        sio_details = SIODetails(file_name)
        logger.info('File {} is determined to be a SIO file.'.format(file_name))
        return SIOReader(sio_details)
    except SarpyIOError:
        return None


#######
#  The actual writing implementation

class SIOWriter(BaseWriter):
    """
    **Changed in version 1.3.0** for writing changes.
    """

    _slots__ = (
        '_file_name', '_file_object', '_in_memory',
        '_data_offset', '_data_written')

    def __init__(
            self,
            file_object: Union[str, BinaryIO],
            sicd_meta: SICDType,
            user_data: Optional[Dict[str, str]] = None,
            check_older_version: bool = False,
            check_existence: bool = True):
        """

        Parameters
        ----------
        file_object : str|BinaryIO
        sicd_meta : SICDType
        user_data : None|Dict[str, str]
        check_older_version : bool
            Try to use an older version (1.1) of the SICD standard, for possible
            application compliance issues?
        check_existence : bool
            Should we check if the given file already exists, and raises an exception if so?
        """

        self._data_written = True
        if isinstance(file_object, str):
            if check_existence and os.path.exists(file_object):
                raise SarpyIOError(
                    'Given file {} already exists,\n\t'
                    'and a new SIO file cannot be created here.'.format(file_object))
            file_object = open(file_object, 'wb')

        if not is_file_like(file_object):
            raise ValueError('file_object requires a file path or BinaryIO object')

        self._file_object = file_object
        if is_real_file(file_object):
            self._file_name = file_object.name
            self._in_memory = False
        else:
            self._file_name = None
            self._in_memory = True

        # choose magic number (with user data) and corresponding endian-ness
        magic_number = 0xFD7F02FF
        endian = SIODetails.ENDIAN[magic_number]

        # define basic image details
        raw_shape = (sicd_meta.ImageData.NumRows, sicd_meta.ImageData.NumCols, 2)
        pixel_type = sicd_meta.ImageData.PixelType
        if pixel_type == 'RE32F_IM32F':
            raw_dtype = numpy.dtype('{}f4'.format(endian))
            element_type = 13
            element_size = 8
            format_function = ComplexFormatFunction(raw_dtype, order='IQ', band_dimension=2)
        elif pixel_type == 'RE16I_IM16I':
            raw_dtype = numpy.dtype('{}i2'.format(endian))
            element_type = 12
            element_size = 4
            format_function = ComplexFormatFunction(raw_dtype, order='IQ', band_dimension=2)
        else:
            raw_dtype = numpy.dtype('{}u1'.format(endian))
            element_type = 11
            element_size = 2
            format_function = AmpLookupFunction(raw_dtype, sicd_meta.ImageData.AmpTable)

        # construct the sio header
        header = numpy.array(
            [raw_shape[0], raw_shape[1], element_type, element_size],
            dtype='>u4')
        # construct the user data - must be {str : str}
        if user_data is None:
            user_data = {}
        uh_args = sicd_meta.get_des_details(check_older_version)
        user_data['SICDMETA'] = sicd_meta.to_xml_string(tag='SICD', urn=uh_args['DESSHTN'])

        # write the initial things to the buffer
        self._file_object.seek(0, os.SEEK_SET)
        self._file_object.write(struct.pack('>I', magic_number))
        self._file_object.write(struct.pack('{}4I'.format(endian), *header))
        self._file_object.write(struct.pack('{}I'.format(endian), len(user_data)))
        # write the user data - name size, name, value size, value
        for name in user_data:
            name_bytes = name.encode('utf-8')
            self._file_object.write(struct.pack('{}I'.format(endian), len(name_bytes)))
            self._file_object.write(struct.pack('{}{}s'.format(endian, len(name_bytes)), name_bytes))
            val_bytes = user_data[name].encode('utf-8')
            self._file_object.write(struct.pack('{}I'.format(endian), len(val_bytes)))
            self._file_object.write(struct.pack('{}{}s'.format(endian, len(val_bytes)), val_bytes))
        self._data_offset = self._file_object.tell()
        # initialize the single data segment
        if self._in_memory:
            underlying_array = numpy.full(raw_shape, fill_value=0, dtype=raw_dtype)
            data_segment = NumpyArraySegment(
                underlying_array, 'complex64', raw_shape[:2], format_function=format_function, mode='w')
            self._data_written = False
        else:
            data_segment = NumpyMemmapSegment(
                self._file_name, self._data_offset, raw_dtype, raw_shape,
                'complex64', raw_shape[:2], format_function=format_function, mode='w', close_file=False)
            self._data_written = True
        BaseWriter.__init__(self, data_segment)

    @property
    def file_name(self) -> Optional[str]:
        """
        None|str: The file name, if feasible.
        """

        return self._file_name

    def flush(self, force: bool = False) -> None:
        BaseWriter.flush(self, force=force)
        if self._data_written:
            return

        if force or self.data_segment[0].check_fully_written(warn=force):
            self._file_object.seek(self._data_offset, os.SEEK_SET)
            self._file_object.write(self.data_segment[0].get_raw_bytes(warn=False))

    def close(self) -> None:
        """
        Completes any necessary final steps.
        """

        if not hasattr(self, '_closed') or self._closed:
            return

        self.flush(force=True)
        BaseWriter.close(self)
        self._file_object = None



"""
This module provide utilities for extracting a suggested name for a SICD.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
import pkgutil
from importlib import import_module
from datetime import datetime

import sarpy._extensions


logger = logging.getLogger(__name__)

###########
# Module variables
_name_functions = []
_parsed_name_functions = False


def register_name_function(name_func):
    """
    Provide a new name function.

    Parameters
    ----------
    name_func : dict

    Returns
    -------
    None
    """

    if not callable(name_func):
        raise TypeError('name_func must be a callable')
    if name_func not in _name_functions:
        _name_functions.append(name_func)


def parse_name_functions():
    """
    Automatically find the viable name functions in the top-level modules.
    """

    global _parsed_name_functions
    if _parsed_name_functions:
        return
    _parsed_name_functions = True

    start_package = 'sarpy.io.complex.naming'
    module = import_module(start_package)
    for details in pkgutil.walk_packages(module.__path__, start_package+'.'):
        _, module_name, is_pkg = details
        if is_pkg:
            # don't bother checking for packages
            continue
        sub_module = import_module(module_name)
        if hasattr(sub_module, 'get_commercial_id'):
            register_name_function(sub_module.get_commercial_id)

    for entry in sarpy._extensions.entry_points(group='sarpy.io.complex.naming.get_commercial_id'):
        register_name_function(entry.load())


def get_sicd_name(the_sicd, product_number=1):
    """
    Gets the suggested name.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType
        The sicd structure.
    product_number : int
        The index of the product from the present file.

    Returns
    -------
    str
    """

    def get_commercial_id():
        commercial_id = None
        for entry in _name_functions:
            commercial_id = entry(collector, cdate_str, cdate_mins, product_number)
            if commercial_id is not None:
                break
        if commercial_id is None:
            return '{0:s}_{1:03d}'.format(the_sicd.CollectionInfo.CoreName, product_number)
        return commercial_id

    def get_vendor_id():
        _time_str = 'HHMMSS' if cdate is None else cdate.strftime('%H%M%S')
        _mode = '{}{}{}'.format(the_sicd.CollectionInfo.RadarMode.get_mode_abbreviation(),
                                the_sicd.Grid.get_resolution_abbreviation(),
                                the_sicd.SCPCOA.SideOfTrack)
        _coords = the_sicd.GeoData.SCP.get_image_center_abbreviation()
        _freq_band = the_sicd.RadarCollection.TxFrequency.get_band_abbreviation()
        _pol = '{}{}'.format(
            the_sicd.RadarCollection.get_polarization_abbreviation(),
            the_sicd.ImageFormation.get_polarization_abbreviation())
        return '_{}_{}_{}_001{}_{}_0101_SPY'.format(_time_str, _mode, _coords, _freq_band, _pol)

    # parse name function, if not already done
    parse_name_functions()

    # extract the common use variables
    if the_sicd.Timeline.CollectStart is None:
        cdate = None
        cdate_str = "DATE"
        cdate_mins = 0
    else:
        start_time = the_sicd.Timeline.CollectStart.astype('datetime64[s]')
        cdate = start_time.astype(datetime)
        cdate_str = cdate.strftime('%d%b%y')
        cdate_mins = cdate.hour * 60 + cdate.minute + cdate.second / 60.

    if the_sicd.CollectionInfo.CollectorName is None:
        collector = 'Unknown'
    else:
        collector = the_sicd.CollectionInfo.CollectorName.strip()

    # noinspection PyBroadException
    try:
        return get_commercial_id() + get_vendor_id()
    except Exception:
        logger.error('Failed to construct suggested name.')
        return None


def get_pass_number(minutes, orbits_per_day):
    """
    Gets appropriately formatted pass number string.

    Parameters
    ----------
    minutes : float
        Minutes elapsed in the day since midnight UTC.
    orbits_per_day : float
        The number of orbits per day, around 15 for vehicles in low earth orbit.

    Returns
    -------
    str
    """

    return '{0:02d}'.format(int(round(minutes*orbits_per_day/1440.)))



"""
Setup module for SarPy.
"""

import os
from setuptools import setup, find_packages

# Get the long description from the README file
here = os.path.abspath(os.path.dirname(__file__))
# Get the long description from the README file
with open(os.path.join(here, 'README.md'), 'r') as f:
    long_description = f.read()


# Get the relevant setup parameters from the package
parameters = {}
with open(os.path.join(here, 'sarpy', '__about__.py'), 'r') as f:
    exec(f.read(), parameters)


def my_package_data():
    def find_dirs(init_dir, start, the_list):
        for root, dirs, files in os.walk(os.path.join(init_dir, start)):
            include_root = False
            for fil in files:
                if os.path.splitext(fil)[1] == '.xsd':
                    include_root = True
                    break
            if include_root:
                root_dir = root.replace('\\', '/')
                rel_dir = root_dir[len(init_dir):]
                if rel_dir[0] == '/':
                    rel_dir = rel_dir[1:]
                if rel_dir[-1] == '/':
                    rel_dir = rel_dir[:-1]
                the_list.append(rel_dir + '/*.xsd')

    package_list = []
    find_dirs('sarpy', 'io/complex/sicd_schema/', package_list)
    find_dirs('sarpy', 'io/phase_history/cphd_schema/', package_list)
    find_dirs('sarpy', 'io/phase_history/crsd_schema/', package_list)
    find_dirs('sarpy', 'io/product/sidd_schema/', package_list)
    find_dirs('sarpy', 'annotation/afrl_rde_schema/', package_list)
    return package_list


def my_test_suite():
    import unittest
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('tests', top_level_dir='.')
    return test_suite


setup(name=parameters['__title__'],
      version=parameters['__version__'],
      description=parameters['__summary__'],
      long_description=long_description,
      long_description_content_type='text/markdown',
      packages=find_packages(exclude=('*tests*', )),
      include_package_data=True,
      package_data={'sarpy': my_package_data()},
      url=parameters['__url__'],
      author=parameters['__author__'],
      author_email=parameters['__email__'],  # The primary POC
      install_requires=['numpy>=1.19.0', 'scipy'],
      zip_safe=False,  # Use of __file__ and __path__ in some code makes it unusable from zip
      test_suite="setup.my_test_suite",
      extras_require={
        "all": ['pillow', 'lxml>=4.1.1', 'matplotlib', 'h5py', 'smart_open[http]', 'pytest>=3.3.2', 'networkx>=2.5', 'shapely>=1.6.4'],
      },
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Programming Language :: Python :: 3.12',
      ],
      platforms=['any'],
      license='MIT')



"""
Create products based on SICD type reader.

For a basic help on the command-line, check

>>> python -m sarpy.utils.create_product --help

"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import argparse
import logging

import sarpy
from sarpy.io.complex.converter import open_complex
from sarpy.processing.ortho_rectify import BivariateSplineMethod, NearestNeighborMethod
from sarpy.processing.sidd.sidd_product_creation import create_detected_image_sidd, \
    create_csi_sidd, create_dynamic_image_sidd
import sarpy.visualization.remap as remap


def _parse_method(method):
    if method.startswith('spline_'):
        return int(method[-1])
    else:
        return 'nearest'


def main(args=None):
    parser = argparse.ArgumentParser(
        description="Create derived product is SIDD format from a SICD type file.",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        'input_file', metavar='input_file',
        help='Path input data file, or directory for radarsat, RCM, or sentinel.\n'
             '* For radarsat or RCM, this can be the product.xml file, or parent directory\n'
             '  of product.xml or metadata/product.xml.\n'
             '* For sentinel, this can be the manifest.safe file, or parent directory of\n'
             '  manifest.safe.\n')
    parser.add_argument(
        'output_directory', metavar='output_directory',
        help='Path to the output directory where the product file(s) will be created.\n'
             'This directory MUST exist.\n'
             '* Depending on the input file, multiple product files may be produced.\n'
             '* The name for the output file(s) will be chosen based on CoreName and\n '
             '  transmit/collect polarization.\n')
    parser.add_argument(
        '-t', '--type', default='detected', choices=['detected', 'csi', 'dynamic'],
        help="The type of derived product.")
    parser.add_argument(
        '-r', '--remap', default=remap.get_remap_names()[0], choices=remap.get_remap_names(),
        help="The pixel value remap function. (default: %(default)s)")
    parser.add_argument(
        '-m', '--method', default='nearest', choices=['nearest', ]+['spline_{}'.format(i) for i in range(1, 6)],
        help="The interpolation method.")
    parser.add_argument(
        '--version', default=2, type=int, choices=[1, 2, 3],
        help="The version of the SIDD standard used.")
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='Verbose (level="INFO") logging?')
    parser.add_argument(
        '-s', '--sicd', action='store_true', help='Include the SICD structure in the SIDD?')

    args = parser.parse_args(args)

    level = 'INFO' if args.verbose else 'WARNING'
    logging.basicConfig(level=level)
    logger = logging.getLogger('sarpy')
    logger.setLevel(level)

    reader = open_complex(args.input_file)
    degree = _parse_method(args.method)

    for i, sicd in enumerate(reader.get_sicds_as_tuple()):
        if isinstance(degree, int):
            ortho_helper = BivariateSplineMethod(reader, index=i, row_order=degree, col_order=degree)
        else:
            ortho_helper = NearestNeighborMethod(reader, index=i)
        if args.type == 'detected':
            create_detected_image_sidd(ortho_helper, args.output_directory,
                                       remap_function=remap.get_registered_remap(args.remap),
                                       version=args.version, include_sicd=args.sicd)
        elif args.type == 'csi':
            create_csi_sidd(ortho_helper, args.output_directory,
                            remap_function=remap.get_registered_remap(args.remap),
                            version=args.version, include_sicd=args.sicd)
        elif args.type == 'dynamic':
            create_dynamic_image_sidd(ortho_helper, args.output_directory,
                                      remap_function=remap.get_registered_remap(args.remap),
                                      version=args.version, include_sicd=args.sicd)
        else:
            raise ValueError('Got unhandled type {}'.format(args.type))


if __name__ == '__main__':
    main()
